package yacs.solver.fdsolver.consistency;

import yacs.domain.*;
import yacs.exceptions.*;
import yacs.net.*;
import yacs.parser.*;
import yacs.solver.*;

import java.util.*;

import org.apache.log4j.*;

/**
 * <p>Einfacher Algorithmus zum Herstellen von Knotenkonsistenz ber finite
 * Domnen nach Marriott und Stuckey (1999, S. 94).</p>
 *
 * <p>Copyright (C) 2005  Wolfgang Runte
 * <br><br>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <br><br>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <br><br>
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 * USA</p>
 *
 * @author Wolfgang Runte (woru@tzi.org)
 * @version YACS 0.1.1
 */
public class NCSolver extends ConsistencySolver {

// ********** private attributs **********

  /** Logger aktivieren: */
  private static Logger logger = Logger.getLogger("yacs.solver.fdsolver.consistency.NCSolver");

// ********** public attributs **********

// ********** private methods **********

  /**
   * Stellt Knotenkonsistenz fr einen primitiven Constraint-Ausdruck her.
   * Liefert <code>true</code>, wenn Werte aus der Domne entfernt wurden,
   * <code>false</code> wenn nicht.
   * @param expr Expression
   * @return boolean
   */
  private boolean nodeConsistentPrimitive(Expression expr) {
    String variable;
    Domain domain;
    DomainElement value;
    boolean removed = false;
    List variablesList = expr.getVariables();
    // Nur auf unre Constraints anwenden:
    if (variablesList.size() == 1) {
      // Variablen und Domnen sichern:
      variable = (String)variablesList.get(0);
      domain = expr.getDomain(variable);
      Iterator it = domain.getElements().iterator();
      // Domne der Variable durchgehen:
      while (it.hasNext()) {
        this.logger.debug("domain: "+domain);
        value = (DomainElement)it.next();
        expr.set(variable, value);
        // Ausdruck mit der gesetzten Variablen auf Erfllbarkeit testen:
        this.logger.debug("expr: "+expr);
        if (expr.satisfiable()) {
          this.logger.debug("satisfiable");
        } else {
          // Wenn kein passender Wert gefunden wurde, den Wert der Variable
          // aus deren Domne entfernen:
          this.logger.info("Entferne Wert '"+value+"' aus Domne von '"+variable+"'.");
          domain.deleteElement(value);
          removed = true;
        }
      }
    }
    return removed;
  }

// ********** public methods **********

  /**
   * Eine Solver der Knotenkonsistenz herstellt, d.h. es werden alle unren
   * Constraints auf konsistente Wertebereiche berprft und selbige ggf. um
   * unzulssige Werte reduziert. Liefert <code>true</code>, wenn
   * Wertebereichseinschrnkungen vorgenommen wurden, <code>false</code> wenn
   * nicht.
   * @param expr Expression
   * @return boolean
   * @throws InconsistencyException
   */
  public boolean nodeSolve(Expression expr) throws InconsistencyException {
    boolean domainModification = false;
    Expression primitiveExpr;
    // Liste der (primitiven) Constraints bzw. Constraint-Ausdrcke holen:
    List primitiveExpressions = expr.getPrimitiveExpressions();
    // (Anm.: Anstatt an dieser Stelle die Methode "getConstraints()" aus der
    // Klasse "ConstraintNet" zu verwenden, hat der obige Aufruf den Vorteil
    // unabhngig von einem bestimmten Constraint-Netz zu sein. Bestehende
    // Konsistenzalgorithmen knnen dadurch z.B. auch von Suchalgorithmen
    // als "look-ahead"-Mechanismus fr bestimmte Teilprobleme ("Expressions")
    // aufgerufen werden.
    for (int i = 0; i < primitiveExpressions.size(); i++) {
      primitiveExpr = (Expression)primitiveExpressions.get(i);
      // Primitives Constraint auf Konsistenz prfen:
      domainModification |= this.nodeConsistentPrimitive(primitiveExpr);
    }
    // Prfen, ob eine Domne leer ist (keine Knotenkonsistenz herstellbar):
    if (expr.hasFalseDomain()) {
      throw new InconsistencyException();
    }
    return domainModification;
  }

  /**
   * Bentigt als Eingabe ein Constraint-Netz. Liefert <code>true</code>, wenn
   * Wertebereichseinschrnkungen vorgenommen wurden, <code>false</code> wenn
   * nicht.
   * @param constraintNet ConstraintNet
   * @return boolean
   * @throws InconsistencyException
   */
  public boolean propagate(ConstraintNet constraintNet) throws InconsistencyException {
    return nodeSolve(constraintNet.getSubproblem());
  }

}
